
import matplotlib.pyplot as plt

from workspace_jam.scripts.requirements import parallel_x2_requirements
from workspace_jam.scripts.requirements import parallel_x3_requirements
from workspace_jam.scripts.requirements import serial_x2_requirements

ci_resolutions = ['low_resolution', 'mid_resolution']  # , 'high_resolution']

requirement_means_parallel_x2, requirement_errors_parallel_x2 = \
    parallel_x2_requirements.parallel_x2_requirements_of_resolutions(ci_resolutions=ci_resolutions)

requirement_means_parallel_x3, requirement_errors_parallel_x3 = \
    parallel_x3_requirements.parallel_x3_requirements_of_resolutions(ci_resolutions=ci_resolutions)

# requirement_means_serial_x2, requirement_errors_serial_x2 = \
#     serial_x2_requirements.serial_x2_requirements_of_resolutions(ci_resolutions=ci_resolutions)

columns = [517, 1034]#, 2068]#, 4136]

plt.figure(figsize=(20, 15))
plt.suptitle('Requirements for CTI Models', fontsize=20)

eb0 = plt.errorbar(x=columns, y=requirement_means_parallel_x2,
                  yerr=[requirement_errors_parallel_x2, requirement_errors_parallel_x2],
                  capsize=10, elinewidth=2, markeredgewidth=5, linestyle=':')
eb0[-1][0].set_linestyle(':')

eb1 = plt.errorbar(x=columns, y=requirement_means_parallel_x3,
                  yerr=[requirement_errors_parallel_x3, requirement_errors_parallel_x3],
                  capsize=10, elinewidth=2, markeredgewidth=5, linestyle=':')
eb1[-1][0].set_linestyle(':')

plt.plot(columns, len(columns) * [1.0], linestyle='-', color='r')
plt.plot(columns, len(columns) * [-1.0], linestyle='-', color='r')

plt.legend(handles=[eb0, eb1], labels=['Parallel x2', 'Parallel x3'], fontsize=25)
plt.xlabel('Number of columns', fontsize=12)
plt.ylabel('Delta Ellipticity / Requirement (1.1e-4)', fontsize=12)
plt.show()


from autocti.charge_injection import ci_frame

def normalization_from_ci_data_resolution(ci_data_resolution):
    return [100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

def parallel_ci_regions_from_ci_data_resolution(ci_data_resolution):
    shape = parallel_shape_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
    return [(0, 30, 51, shape[1]-20), (330, 360, 51, shape[1]-20),
              (660, 690, 51, shape[1]-20), (990, 1020, 51, shape[1]-20),
              (1320, 1350, 51, shape[1]-20), (1650, 1680, 51, shape[1]-20),
              (1980, 2010, 51, shape[1]-20)]


def serial_ci_regions_from_ci_data_resolution(ci_data_resolution):
    shape = serial_shape_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
    return [(0, shape[0], 51, shape[1]-20)]

def parallel_shape_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution is 'high_resolution':
        return (2316, 2119)
    elif ci_data_resolution is 'mid_resolution':
        return (2316, 1034)
    elif ci_data_resolution is 'low_resolution':
        return (2316, 517)

def parallel_frame_geometry_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution is 'high_resolution':
        return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))

    elif ci_data_resolution is 'mid_resolution':
        return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 1014)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 1014, 1034)))

    elif ci_data_resolution is 'low_resolution':
        return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 497)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 497, 517)))

def serial_shape_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution is 'high_resolution':
        return (2066, 2119)

    elif ci_data_resolution is 'mid_resolution':
        return (1033, 2119)

    elif ci_data_resolution is 'low_resolution':
        return (517, 2119)

def serial_frame_geometry_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution is 'high_resolution':
        return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))


    elif ci_data_resolution is 'mid_resolution':
        return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((1138, 1158, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 1158, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 1138, 2099, 2119)))


    elif ci_data_resolution is 'low_resolution':
        return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                                parallel_overscan=ci_frame.Region((559, 579, 51, 2099)),
                                                serial_prescan=ci_frame.Region((0, 579, 0, 51)),
                                                serial_overscan=ci_frame.Region((0, 559, 2099, 2119)))

def parallel_and_serial_shape():
    return (2316, 2119)

def parallel_and_serial_frame_geometry():
    return ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))

def parallel_and_serial_ci_regions_from_ci_data_resolution():
    shape = parallel_and_serial_shape()
    return [(0, 30, 51, shape[1]-20), (330, 360, 51, shape[1]-20),
              (660, 690, 51, shape[1]-20), (990, 1020, 51, shape[1]-20),
              (1320, 1350, 51, shape[1]-20), (1650, 1680, 51, shape[1]-20),
              (1980, 2010, 51, shape[1]-20)]

def charge_line_shape_and_frame_geometry_from_direction(direction):

    if direction is 'parallel':

        shape = (2316, 1)
        frame_geometry = ci_frame.FrameGeometry.euclid_parallel_line()

        return shape, frame_geometry

    elif direction is 'serial':

        shape = (1, 2119)
        frame_geometry = ci_frame.FrameGeometry.euclid_serial_line()

        return shape, frame_geometry


from matplotlib import pyplot as plt

from autocti.charge_injection.plotters import fit_plotters
from autocti.plotters import plotter_util
from autocti.data import util

def plot_ci_fit_for_phase(
        fits, during_analysis,
        extract_array_from_mask,
        should_plot_all_at_end_png,
        should_plot_all_at_end_fits,
        should_plot_as_subplot,
        should_plot_residual_maps_subplot,
        should_plot_chi_squared_maps_subplot,
        should_plot_image,
        should_plot_noise_map,
        should_plot_signal_to_noise_map,
        should_plot_ci_pre_cti,
        should_plot_ci_post_cti,
        should_plot_residual_map,
        should_plot_chi_squared_map,
        should_plot_noise_scaling_maps,
        should_plot_parallel_front_edge_line,
        should_plot_parallel_trails_line,
        should_plot_serial_front_edge_line,
        should_plot_serial_trails_line,
        visualize_path=None):

    plot_ci_fit_arrays_for_phase(
        fits=fits, during_analysis=during_analysis,
        extract_array_from_mask=extract_array_from_mask,
        should_plot_all_at_end_png=should_plot_all_at_end_png,
        should_plot_all_at_end_fits=should_plot_all_at_end_fits,
        should_plot_as_subplot=should_plot_as_subplot,
        should_plot_residual_maps_subplot=should_plot_residual_maps_subplot,
        should_plot_chi_squared_maps_subplot=should_plot_chi_squared_maps_subplot,
        should_plot_image=should_plot_image,
        should_plot_noise_map=should_plot_noise_map,
        should_plot_signal_to_noise_map=should_plot_signal_to_noise_map,
        should_plot_ci_pre_cti=should_plot_ci_pre_cti,
        should_plot_ci_post_cti=should_plot_ci_post_cti,
        should_plot_noise_scaling_maps=should_plot_noise_scaling_maps,
        should_plot_residual_map=should_plot_residual_map,
        should_plot_chi_squared_map=should_plot_chi_squared_map,
        visualize_path=visualize_path)

    plot_ci_fit_lines_for_phase(
        fits=fits, during_analysis=during_analysis,
        should_plot_all_at_end_png=should_plot_all_at_end_png,
        should_plot_all_at_end_fits=should_plot_all_at_end_fits,
        should_plot_as_subplot=should_plot_as_subplot,
        should_plot_residual_maps_subplot=should_plot_residual_maps_subplot,
        should_plot_chi_squared_maps_subplot=should_plot_chi_squared_maps_subplot,
        should_plot_image=should_plot_image,
        should_plot_noise_map=should_plot_noise_map,
        should_plot_signal_to_noise_map=should_plot_signal_to_noise_map,
        should_plot_ci_pre_cti=should_plot_ci_pre_cti,
        should_plot_ci_post_cti=should_plot_ci_post_cti,
        should_plot_residual_map=should_plot_residual_map,
        should_plot_chi_squared_map=should_plot_chi_squared_map,
        should_plot_parallel_front_edge_line=should_plot_parallel_front_edge_line,
        should_plot_parallel_trails_line=should_plot_parallel_trails_line,
        should_plot_serial_front_edge_line=should_plot_serial_front_edge_line,
        should_plot_serial_trails_line=should_plot_serial_trails_line,
        visualize_path=visualize_path)

def plot_ci_fit_arrays_for_phase(
        fits, during_analysis,
        extract_array_from_mask,
        should_plot_all_at_end_png,
        should_plot_all_at_end_fits,
        should_plot_as_subplot,
        should_plot_residual_maps_subplot,
        should_plot_chi_squared_maps_subplot,
        should_plot_image,
        should_plot_noise_map,
        should_plot_signal_to_noise_map,
        should_plot_ci_pre_cti,
        should_plot_ci_post_cti,
        should_plot_residual_map,
        should_plot_chi_squared_map,
        should_plot_noise_scaling_maps,
        visualize_path=None):

    for fit in fits:

        normalization = fit.ci_data_fit.ci_pattern.normalization
        output_path = visualize_path + '/' + 'ci_image_' + str(int(normalization)) + '/arrays/'
        util.make_path_if_does_not_exist(path=output_path + 'fits/')

        if should_plot_as_subplot:

            plot_fit_subplot(
                fit=fit, extract_array_from_mask=extract_array_from_mask,
                output_path=output_path, output_format='png')

        plot_fit_individuals(
            fit=fit,
            extract_array_from_mask=extract_array_from_mask,
            should_plot_image=should_plot_image,
            should_plot_noise_map=should_plot_noise_map,
            should_plot_signal_to_noise_map=should_plot_signal_to_noise_map,
            should_plot_ci_pre_cti=should_plot_ci_pre_cti,
            should_plot_ci_post_cti=should_plot_ci_post_cti,
            should_plot_residual_map=should_plot_residual_map,
            should_plot_chi_squared_map=should_plot_chi_squared_map,
            should_plot_noise_scaling_maps=should_plot_noise_scaling_maps,
            output_path=output_path, output_format='png')

        if not during_analysis:

            if should_plot_all_at_end_png:

                plot_fit_individuals(
                    fit=fit,
                    extract_array_from_mask=extract_array_from_mask,
                    should_plot_image=True,
                    should_plot_noise_map=True,
                    should_plot_signal_to_noise_map=True,
                    should_plot_ci_pre_cti=True,
                    should_plot_ci_post_cti=True,
                    should_plot_residual_map=True,
                    should_plot_chi_squared_map=True,
                    should_plot_noise_scaling_maps=True,
                    output_path=output_path, output_format='png')

            if should_plot_all_at_end_fits:

                plot_fit_individuals(
                    fit=fit,
                    extract_array_from_mask=extract_array_from_mask,
                    should_plot_image=True,
                    should_plot_noise_map=True,
                    should_plot_signal_to_noise_map=True,
                    should_plot_ci_pre_cti=True,
                    should_plot_ci_post_cti=True,
                    should_plot_residual_map=True,
                    should_plot_chi_squared_map=True,
                    should_plot_noise_scaling_maps=True,
                    output_path="{}/fits/".format(output_path), output_format='fits')

        output_path = visualize_path + '/'

        if should_plot_residual_maps_subplot:

            plot_fit_residual_maps_subplot(
                fits=fits, extract_array_from_mask=extract_array_from_mask,
                output_path=output_path, output_format='png')

        if should_plot_chi_squared_maps_subplot:

            plot_fit_chi_squared_maps_subplot(
                fits=fits, extract_array_from_mask=extract_array_from_mask,
                output_path=output_path, output_format='png')

def plot_ci_fit_lines_for_phase(
        fits, during_analysis,
        should_plot_all_at_end_png,
        should_plot_all_at_end_fits,
        should_plot_as_subplot,
        should_plot_residual_maps_subplot,
        should_plot_chi_squared_maps_subplot,
        should_plot_image,
        should_plot_noise_map,
        should_plot_signal_to_noise_map,
        should_plot_ci_pre_cti,
        should_plot_ci_post_cti,
        should_plot_residual_map,
        should_plot_chi_squared_map,
        should_plot_parallel_front_edge_line,
        should_plot_parallel_trails_line,
        should_plot_serial_front_edge_line,
        should_plot_serial_trails_line,
        visualize_path=None):

    line_regions = plotter_util.line_regions_from_should_plots(
        should_plot_parallel_front_edge_line=should_plot_parallel_front_edge_line,
        should_plot_parallel_trails_line=should_plot_parallel_trails_line,
        should_plot_serial_front_edge_line=should_plot_serial_front_edge_line,
        should_plot_serial_trails_line=should_plot_serial_trails_line)

    for line_region in line_regions:

        for fit in fits:

            normalization = fit.ci_data_fit.ci_pattern.normalization
            output_path = visualize_path + '/' + 'ci_image_' + str(int(normalization)) + '/' + line_region + '/'
            util.make_path_if_does_not_exist(path=output_path + 'fits/')

            if should_plot_as_subplot:

                plot_fit_line_subplot(
                    fit=fit, line_region=line_region,
                    output_path=output_path, output_format='png')

            plot_fit_line_individuals(
                fit=fit, line_region=line_region,
                should_plot_image=should_plot_image,
                should_plot_noise_map=should_plot_noise_map,
                should_plot_signal_to_noise_map=should_plot_signal_to_noise_map,
                should_plot_ci_pre_cti=should_plot_ci_pre_cti,
                should_plot_ci_post_cti=should_plot_ci_post_cti,
                should_plot_residual_map=should_plot_residual_map,
                should_plot_chi_squared_map=should_plot_chi_squared_map,
                output_path=output_path, output_format='png')

            if not during_analysis:

                if should_plot_all_at_end_png:

                    plot_fit_line_individuals(
                        fit=fit, line_region=line_region,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_signal_to_noise_map=True,
                        should_plot_ci_pre_cti=True,
                        should_plot_ci_post_cti=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        output_path=output_path, output_format='png')

                if should_plot_all_at_end_fits:

                    plot_fit_line_individuals(
                        fit=fit, line_region=line_region,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_signal_to_noise_map=True,
                        should_plot_ci_pre_cti=True,
                        should_plot_ci_post_cti=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        output_path=output_path, output_format='fits')

        output_path = visualize_path + '/'  + line_region + '_'

        if should_plot_residual_maps_subplot:

            plot_fit_residual_maps_lines_subplot(
                fits=fits, line_region=line_region,
                output_path=output_path, output_format='png')

        if should_plot_chi_squared_maps_subplot:

            plot_fit_chi_squared_maps_lines_subplot(
                fits=fits, line_region=line_region,
                output_path=output_path, output_format='png')


def plot_fit_subplot(
        fit, extract_array_from_mask=False,
        figsize=None, aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
        output_path=None, output_filename='ci_fit', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=9)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    fit_plotters.plot_image(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 2)

    fit_plotters.plot_noise_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 3)

    fit_plotters.plot_signal_to_noise_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 4)

    fit_plotters.plot_ci_pre_cti(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 5)

    fit_plotters.plot_ci_post_cti(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 7)

    fit_plotters.plot_residual_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 8)

    fit_plotters.plot_chi_squared_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad, cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_residual_maps_subplot(
        fits, extract_array_from_mask=False,
        figsize=None, aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
        output_path=None, output_filename='ci_fits_residual_maps', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=len(fits))

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    
    for index, fit in enumerate(fits):
    
        plt.subplot(rows, columns, index+1)

        fit_plotters.plot_residual_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            as_subplot=True,
            figsize=figsize, aspect=aspect,
            cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
            cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
            cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
            titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_chi_squared_maps_subplot(
        fits, extract_array_from_mask=False,
        figsize=None, aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
        output_path=None, output_filename='ci_fits_chi_squared_maps', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including chi_squared_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=len(fits))

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for index, fit in enumerate(fits):

        plt.subplot(rows, columns, index+1)

        fit_plotters.plot_chi_squared_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            as_subplot=True,
            figsize=figsize, aspect=aspect,
            cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
            cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
            cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
            titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_individuals(
        fit, extract_array_from_mask=False,
        should_plot_image=False,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_ci_pre_cti=False,
        should_plot_ci_post_cti=False,
        should_plot_residual_map=False,
        should_plot_chi_squared_map=False,
        should_plot_noise_scaling_maps=False,
        output_path=None, output_format='show'):

    if should_plot_image:
        fit_plotters.plot_image(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_noise_map:
        fit_plotters.plot_noise_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_signal_to_noise_map:
        fit_plotters.plot_signal_to_noise_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_ci_pre_cti:
        fit_plotters.plot_ci_pre_cti(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_ci_post_cti:
        fit_plotters.plot_ci_post_cti(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_residual_map:
        fit_plotters.plot_residual_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_chi_squared_map:
        fit_plotters.plot_chi_squared_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_noise_scaling_maps and hasattr(fit, 'noise_scaling_maps'):

        fit_plotters.plot_noise_scaling_maps(
            fit_hyper=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)


def plot_fit_line_subplot(
        fit, line_region,
        figsize=None,
        titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_filename='ci_fit_line', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=9)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    fit_plotters.plot_image_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 2)

    fit_plotters.plot_noise_map_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 3)

    fit_plotters.plot_signal_to_noise_map_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 4)

    fit_plotters.plot_ci_pre_cti_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 5)

    fit_plotters.plot_ci_post_cti_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 7)

    fit_plotters.plot_residual_map_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 8)

    fit_plotters.plot_chi_squared_map_line(
        fit=fit, line_region=line_region,  mask=fit.mask, as_subplot=True,
       figsize=figsize,
    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_residual_maps_lines_subplot(
        fits, line_region,
        figsize=None,
        titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_filename='ci_fits_residual_maps_lines', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, residual_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=len(fits))

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for index, fit in enumerate(fits):
        plt.subplot(rows, columns, index + 1)

        fit_plotters.plot_residual_map_line(
            fit=fit, mask=fit.mask, line_region=line_region,
            as_subplot=True,
            titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_chi_squared_maps_lines_subplot(
        fits, line_region,
        figsize=None,
        titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_filename='ci_fits_chi_squared_maps_lines', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including chi_squared_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=len(fits))

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for index, fit in enumerate(fits):
        
        plt.subplot(rows, columns, index + 1)

        fit_plotters.plot_chi_squared_map_line(
            fit=fit, mask=fit.mask, line_region=line_region,
            as_subplot=True,
            titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_line_individuals(
        fit, line_region,
        should_plot_image=False,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_ci_pre_cti=False,
        should_plot_ci_post_cti=False,
        should_plot_residual_map=False,
        should_plot_chi_squared_map=False,
        output_path=None, output_format='show'):

    if should_plot_image:
        fit_plotters.plot_image_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)

    if should_plot_noise_map:
        fit_plotters.plot_noise_map_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)

    if should_plot_signal_to_noise_map:
        fit_plotters.plot_signal_to_noise_map_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)

    if should_plot_ci_pre_cti:
        fit_plotters.plot_ci_pre_cti_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)

    if should_plot_ci_post_cti:
        fit_plotters.plot_ci_post_cti_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)

    if should_plot_residual_map:
        fit_plotters.plot_residual_map_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)

    if should_plot_chi_squared_map:
        fit_plotters.plot_chi_squared_map_line(
            fit=fit, line_region=line_region, mask=fit.mask,
            output_path=output_path, output_format=output_format)



import numpy as np
def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.

    values, weights -- Numpy ndarrays with the same shape.
    """
    values = np.asarray(values)
    weights = np.asarray(weights)
    average = np.average(values, weights=weights)
    # Fast and numerically precise:
    variance = np.average((values-average)**2, weights=weights)
    return (average, np.sqrt(variance))


from autofit import conf
from autofit import aggregator
from autocti.model import arctic_params as ap
import os
import numpy as np

from workspace_jam.scripts.requirements.tools import weighted_avg_and_std

def serial_x2_requirements_of_resolutions(ci_resolutions):

    # Setup the path to the workspace, using a relative directory name.
    workspace_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
    output_path = workspace_path + '../../outputs/PyAutoCTI/'

    # Use this path to explicitly set the config path and output path.
    conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=output_path)

    input_serial_species_0 = ap.Species(trap_density=0.13, trap_lifetime=1.25)
    input_serial_species_1 = ap.Species(trap_density=0.25, trap_lifetime=4.4)
    input_arctic_params = ap.ArcticParams(serial_species=[input_serial_species_0, input_serial_species_1])
    input_delta_ellipticity = input_arctic_params.delta_ellipticity

    pipeline = 'pipeline_normal__serial_x2'
    phase = 'phase_1_serial_x2'

    ci_image = 'ci_images_non_uniform_cosmic_rays'
    ci_model = 'serial_x2'
    ci_setting = 'settings_cr_p10s10d3'

    requirement_means_of_resolutions = []
    requirement_stds_of_resolutions = []
    requirement_errors_of_resolutions= []

    for ci_resolution in ci_resolutions:

        agg = aggregator.Aggregator(directory=output_path + ci_image + '/' + ci_model + '/' + ci_resolution + '/' + pipeline + '/' + phase + '/' + ci_setting)
        optimizer = agg.optimizers_with(pipeline=pipeline, phase=phase)[0]

        weights = []
        requirements = []

        for sample_index in range(optimizer.total_samples):

            weight = optimizer.sample_weight_from_sample_index(sample_index=sample_index)

            if weight > 1.0e-4:

                params = optimizer.sample_model_parameters_from_sample_index(sample_index=sample_index)
                params.append(0.0)
                params.append(0.0)
                params[1], params[2] = params[2], params[1]
                instance = optimizer.variable.instance_from_physical_vector(physical_vector=params)
                print(instance.serial_species[0].trap_density)
                print(instance.serial_species[1].trap_density)
                print(instance.serial_species[0].trap_lifetime)
                print(instance.serial_species[1].trap_lifetime)
                stop
        #        instance = optimizer.sample_model_instance_from_sample_index(sample_index=sample_index)
                params = ap.ArcticParams(serial_species=instance.serial_species)
                requirement = (params.delta_ellipticity - input_delta_ellipticity) / 1.1e-4
                weights.append(weight)
                requirements.append(requirement)

        requirement_mean, requirement_std = weighted_avg_and_std(values=requirements, weights=weights)

        requirement_means_of_resolutions.append(requirement_mean)
        requirement_stds_of_resolutions.append(requirement_std)
        requirement_errors_of_resolutions.append(requirement_std / 2.0)

    return requirement_means_of_resolutions, requirement_errors_of_resolutions



from autofit import conf
from autofit import aggregator
import os

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=workspace_path + 'output')

pipeline = 'pipeline_init__parallel_x2 + pipeline_normal__parallel_x2'
phase = 'phase_1_parallel_x2'



agg = aggregator.Aggregator(directory=workspace_path + 'output/ci_images_uniform/parallel_x2/high_resolution/')

opt = agg.optimizers_with(pipeline=pipeline, phase=phase)

print(opt)


import os

from autofit import conf
from autofit.tools import path_util
from autocti.data import mask as msk
from autocti.charge_injection import ci_pattern
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_fit

from autocti.pipeline import phase as ph
from autocti.model import arctic_params
from autocti.model import arctic_settings

from workspace_jam.tools.data_makers import tools

import numpy as np

from multiprocessing import Pool

# Welcome to the pipeline runner. This tool allows you to CTI calibration data on strong lenses, and pass it to
# pipelines for a PyAutoCTI analysis. To show you around, we'll load up some example data and run it through some of
# the example pipelines that come distributed with PyAutoCTI.

# The runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out data-set names
# and pipelines to perform different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# The pipeline runner is fairly self explanatory. Make sure to checkout the pipelines in the
#  workspace/pipelines/examples/ folder - they come with detailed descriptions of what they do. I hope that you'll
# expand on them for your own personal scientific needs

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=workspace_path + 'output')

# It is convenient to specify the image type and model used to simulate that image as strings, so that if the
# pipeline is applied to multiple images we don't have to change all of the path entries in the
# load_ci_data_from_fits function below.

ci_data_type = 'ci_images_uniform_no_noise' # Charge injection data consisting of 2 images with uniform injections.
ci_data_model = 'serial_x3' # Shows the data was creating using a serial CTI model with one species.
ci_data_resolution = 'low_resolution' # The resolution of the image.

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/ci_images_uniform/serial_x3_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

# This tools loads the correct shape of the image and frame geometry, for the ci_data_resolution string.
shape, frame_geometry = \
    tools.serial_shape_and_frame_geometry_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)

# The charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, shape[0], 51, shape[1]-20)]

# The normalization of the charge injection for each image.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# Create the charge injection pattern objects used for this pipeline.
patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

# There are 8 images to load, which is what the normalizations list tells us. To load them, its easiest to create a
# 'data' list and then iterate over a for loop, to append each set of data to the list of data we pass to the pipeline
# when we run it.

datas = []

for index, normalization in enumerate(normalizations):

    datas.append(ci_data.ci_data_from_fits(
                 frame_geometry=frame_geometry, ci_pattern=patterns[index],
                 image_path=ci_data_path + 'image_' + str(int(normalization)) + '.fits',
                 ci_pre_cti_path=ci_data_path +'/ci_pre_cti_' + str(int(normalization)) + '.fits',
                 noise_map_from_single_value=4.0))

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For serial ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
serial_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                        charge_injection_mode=False, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(serial=serial_cti_settings)

serial_species_0 = arctic_params.Species(trap_density=0.01, trap_lifetime=0.8)
serial_species_1 = arctic_params.Species(trap_density=0.03, trap_lifetime=4.0)
serial_species_2 = arctic_params.Species(trap_density=0.9, trap_lifetime=20.0)

serial_ccd = arctic_params.CCD(well_notch_depth=1.0e-4, well_fill_alpha=1.0, well_fill_beta=0.58)
cti_params = arctic_params.ArcticParams(serial_species=[serial_species_0, serial_species_1, serial_species_2],
                                        serial_ccd=serial_ccd)


mask_function=msk.Mask.empty_for_shape
masks = list(map(lambda data: mask_function(shape=data.image.shape), datas))
phase = ph.SerialPhase(rows=(0, 10), phase_name='test')
ci_datas_fit = [phase.extract_ci_data(data=data, mask=mask) for data, mask in zip(datas, masks)]

fits = list(map(lambda ci_data_fit:
                ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=cti_params, cti_settings=cti_settings),
                ci_datas_fit))

print(list(map(lambda fit : np.max(fit.residual_map), fits)))
print(list(map(lambda fit : fit.likelihood, fits)))


import numpy as np

from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_fit
from autocti.charge_injection import ci_hyper
from autocti.charge_injection.plotters import ci_fit_plotters
from autocti.data import mask as msk
from autocti.model import arctic_params
from autocti.model import arctic_settings
from test.unit.mock.mock import MockPattern, MockCIFrame

from test.fixtures import make_plot_patch
import os
import pytest

@pytest.fixture(name='ci_fit_plotter_path')
def make_ci_data_plotter_setup():
    return "{}/../../test_files/plotting/ci_fit/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask.empty_for_shape(shape=(2, 2))


@pytest.fixture(name='image')
def make_image():
    return np.ones((2, 2))


@pytest.fixture(name='noise_map')
def make_noise_map():
    return 2.0 * np.ones((2, 2))


@pytest.fixture(name='ci_pre_cti')
def make_ci_pre_cti():
    return 3.0 * np.ones((2, 2))

@pytest.fixture(name='ci_frame')
def make_ci_frame():
    return MockCIFrame(value=3.0)

@pytest.fixture(name='ci_data_fit')
def make_ci_data_fit(image, noise_map, mask, ci_pre_cti):
    return ci_data.MaskedCIData(image=image, noise_map=noise_map, ci_pre_cti=ci_pre_cti, mask=mask,
                                ci_pattern=MockPattern(), ci_frame=MockCIFrame(value=3.0))


@pytest.fixture(name="cti_settings")
def make_cti_settings():
    parallel_settings = arctic_settings.Settings(well_depth=0, niter=1, express=1, n_levels=2000)
    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)


@pytest.fixture(name="cti_params")
def make_cti_params():
    parallel_1_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
    return arctic_params.ArcticParams(parallel_species=parallel_1_species)


@pytest.fixture(name="fit")
def make_fit(ci_data_fit, cti_params, cti_settings):
    return ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=cti_params,
                                                               cti_settings=cti_settings)



def test__ci_fit_subplot_is_output(fit, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_fit_subplot(fit=fit, extract_array_from_mask=True,
                                     cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'ci_fit.png' in plot_patch.paths

def test__ci_fit_residual_maps_subplot_is_output(fit, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_fit_residual_maps_subplot(fits=[fit], extract_array_from_mask=True,
                                     cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'ci_fits_residual_maps.png' in plot_patch.paths

    ci_fit_plotters.plot_fit_residual_maps_lines_subplot(fits=[fit], line_region='parallel_front_edge',
                                     output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'ci_fits_residual_maps_lines.png' in plot_patch.paths


def test__ci_fit_chi_squareds_subplot_is_output(fit, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_fit_chi_squared_maps_subplot(fits=[fit], extract_array_from_mask=True,
                                     cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'ci_fits_chi_squared_maps.png' in plot_patch.paths

    ci_fit_plotters.plot_fit_chi_squared_maps_lines_subplot(fits=[fit], line_region='parallel_front_edge',
                                     output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'ci_fits_chi_squared_maps_lines.png' in plot_patch.paths

def test__fit_individuals__depedent_on_input(fit, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_fit_individuals(
        fit=fit,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_ci_pre_cti=True,
        should_plot_ci_post_cti=True,
        should_plot_chi_squared_map=True,
        output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'fit_image.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_ci_pre_cti.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_ci_post_cti.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths


def test__fit_individuals_line__depedent_on_input(fit, ci_frame, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_fit_line_individuals(
        fit=fit, line_region='parallel_front_edge',
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_ci_pre_cti=True,
        should_plot_ci_post_cti=True,
        should_plot_chi_squared_map=True,
        output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'fit_image_line.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_noise_map_line.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_signal_to_noise_map_line.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_ci_pre_cti_line.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_ci_post_cti_line.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_residual_map_line.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_chi_squared_map_line.png' in plot_patch.paths


def test__plot_ci_fit_for_phase(fit, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_ci_fit_for_phase(fits=[fit], during_analysis=False,
                                          extract_array_from_mask=True,
                                          should_plot_all_at_end_png=False,
                                          should_plot_all_at_end_fits=False,
                                          should_plot_as_subplot=True,
                                          should_plot_residual_maps_subplot=True,
                                          should_plot_chi_squared_maps_subplot=False,
                                          should_plot_image=True,
                                          should_plot_noise_map=False,
                                          should_plot_ci_pre_cti=True,
                                          should_plot_signal_to_noise_map=False,
                                          should_plot_ci_post_cti=False,
                                          should_plot_residual_map=True,
                                          should_plot_chi_squared_map=False,
                                          should_plot_noise_scaling_maps=False,
                                          should_plot_parallel_front_edge_line=True,
                                          should_plot_parallel_trails_line=False,
                                          should_plot_serial_front_edge_line=True,
                                          should_plot_serial_trails_line=False,
                                          visualize_path=ci_fit_plotter_path)

    assert ci_fit_plotter_path + '/ci_image_10/arrays/ci_fit.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/arrays/fit_image.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/arrays/fit_noise_map.png' not in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/arrays/fit_ci_pre_cti.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/arrays/fit_signal_to_noise_map.png' not in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/arrays/fit_residual_map.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/arrays/fit_chi_squared_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/ci_fit_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/fit_image_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/fit_noise_map_line.png' not in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/fit_ci_pre_cti_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/fit_signal_to_noise_map_line.png' not in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/fit_residual_map_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/parallel_front_edge/fit_chi_squared_map_line.png' not in plot_patch.paths

    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/ci_fit_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/fit_image_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/fit_noise_map_line.png' not in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/fit_ci_pre_cti_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/fit_signal_to_noise_map_line.png' not in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/fit_residual_map_line.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_image_10/serial_front_edge/fit_chi_squared_map_line.png' not in plot_patch.paths

    assert ci_fit_plotter_path + '/ci_fits_residual_maps.png' in plot_patch.paths
    assert ci_fit_plotter_path + '/ci_fits_chi_sqaured_maps.png' not in plot_patch.paths


@pytest.fixture(name='ci_data_fit_hyper')
def make_ci_data_fit_hyper(image, noise_map, mask, ci_pre_cti):
    return ci_data.MaskedCIHyperData(image=image, noise_map=noise_map, ci_pre_cti=ci_pre_cti, mask=mask,
                                ci_pattern=MockPattern(), ci_frame=MockCIFrame(value=3.0),
                                     noise_scaling_maps=[np.ones((2,2)), 2.0*np.ones((2,2))])

@pytest.fixture(name='hyper_noise_scalars')
def make_hyper_noise_scalars():
    return [ci_hyper.CIHyperNoiseScalar(scale_factor=1.0), ci_hyper.CIHyperNoiseScalar(scale_factor=2.0)]

@pytest.fixture(name="fit_hyper")
def make_fit_hyper(ci_data_fit_hyper, cti_params, cti_settings, hyper_noise_scalars):
    return ci_fit.CIHyperFit(masked_hyper_ci_data=ci_data_fit_hyper, cti_params=cti_params, cti_settings=cti_settings,
                             hyper_noise_scalars=hyper_noise_scalars)

def test__fit_individuals__fit_hyper_plots_noise_scaling_maps(fit, fit_hyper, ci_fit_plotter_path, plot_patch):

    ci_fit_plotters.plot_fit_individuals(
        fit=fit,
        should_plot_noise_scaling_maps=True,
        output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'fit_noise_scaling_maps.png' not in plot_patch.paths

    ci_fit_plotters.plot_fit_individuals(
        fit=fit_hyper,
        should_plot_noise_scaling_maps=True,
        output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'fit_noise_scaling_maps.png' in plot_patch.paths


from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autofit.mapper import prior_model
from autocti.data import mask as msk
from autocti.pipeline import pipeline as pl
from autocti.pipeline import phase as ph
from autocti.model import arctic_params

# In this pipeline, we'll perform a basic analysis which fits a single serial trap species to a set of charge
# injection imaging data. This will include a hyper-phase which scales the noise in the analysis, to prevent
# over-fitting the highest S/N charge injection images. The pipeline uses three phases:

# Phase 1) Fit a small section (the top 10 rows of every charge injection) using a serial CTI model
#          with 1 trap species and a model for the serial CCD volume filling parameters.

# Phase 2) Use the best-fit model from phase 1 to scale the noise of each image, to ensure that the higher and
#          lower S/N images are weighted more equally in their contribution to the likelihood.

# Phase 3) Refit the phase 1 model, using priors initialized from the results of phase 1 and the scaled noise-map
#          computed in phase 2.

def make_pipeline(
        phase_folders=None, tag_phases=True,
        mask_function=msk.Mask.empty_for_shape,
        serial_front_edge_mask_columns=None, serial_trails_mask_columns=None, serial_total_density_range=None,
        cosmic_ray_parallel_buffer=None, cosmic_ray_serial_buffer=None, cosmic_ray_diagonal_buffer=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = 'pipeline_normal__serial_x1'

    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the data with a one species serial CTI model and serial CCD filling model. In this
    # phase we will:

    # 1) Use the complete charge injection image, as opposed to extracting a sub-set of rows.
    # 3) Initialize the priors on the serial CTI model from the results of phase 1 of the initialize pipeline.

    class SerialPhase(ph.SerialPhase):

        def pass_priors(self, results):

            self.serial_species = results.from_phase('phase_1_serial_x1').variable.serial_species
            self.serial_ccd = results.from_phase('phase_1_serial_x1').variable.serial_ccd
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = SerialPhase(
        phase_name='phase_1_serial_x1', phase_folders=phase_folders, tag_phases=tag_phases,
        serial_species=[prior_model.PriorModel(arctic_params.Species)],
        serial_ccd=arctic_params.CCD,
        mask_function=mask_function,
        serial_front_edge_mask_columns=serial_front_edge_mask_columns,
        serial_trails_mask_columns=serial_trails_mask_columns,
        serial_total_density_range=serial_total_density_range,
        cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer, cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
        cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer,
        optimizer_class=nl.MultiNest)

    # For the final CTI model, constant efficiency mode has a tendancy to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase1.optimizer.const_efficiency_mode = False
    phase1.optimizer.n_live_points = 50
    phase1.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(pipeline_name, phase1)


from autofit.tools import path_util
from autocti.data import util
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern

from autocti.model import arctic_settings
from autocti.model import arctic_params

from workspace_jam.tools.data_makers import tools

import os

# This tool allows one to make simulated charge injection imaging data-sets for calibrating parallel charge transfer
# inefficiency, which can be used to test example pipelines and investigate CTI modeling on data-sets where the
# 'true' answer is known.

# The 'ci_data_type', 'ci_data_model' and 'ci_data_resolution' determine the directory the output data folder, e.g:

# The image will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_image.fits'.
# The noise-map will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_noise_map.fits'.
# The pre cti ci image will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_pre_cti.fits'.

# Lets setup the relative path to the workspace, so we can output the data in the 'data' folder..
workspace_path = '{}/../../../../'.format(os.path.dirname(os.path.realpath(__file__)))

ci_data_type = 'ci_images_uniform'
ci_data_model = 'parallel_x2_poisson'
ci_data_resolution = 'low_resolution'

# Create the path where the data will be output, which in this case is
# '/workspace/data/ci_images_uniform/parallel_x2_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

# These tools loads the correct shape, frame geometry, normalizations and charge injection regions for parallel CTI.
shape = tools.parallel_shape_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
frame_geometry = tools.parallel_frame_geometry_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
normalizations = tools.normalization_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
ci_regions = tools.parallel_ci_regions_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For parallel ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
parallel_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                        charge_injection_mode=True, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(parallel=parallel_cti_settings)

# The CTI model parameters of arCTIc, which includes each trap species density / lifetime and the CCD properties for
# parallel charge transfer.
parallel_species_0 = arctic_params.Species(trap_density=0.13, trap_lifetime=1.25)
parallel_species_1 = arctic_params.Species(trap_density=0.25, trap_lifetime=4.4)
parallel_species = arctic_params.Species.poisson_species(species=[parallel_species_0, parallel_species_1], shape=shape,
                                                         seed=1)

parallel_ccd = arctic_params.CCD(well_notch_depth=1.0e-4, well_fill_beta=0.58, well_fill_alpha=1.0, well_fill_gamma=0.0)

cti_params = arctic_params.ArcticParams(parallel_species=parallel_species, parallel_ccd=parallel_ccd)

# Use the ci normalizations and regions to create the ci pattern of every image that is to be simulated.
ci_patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

# Use the simulate ci patterns to generate the pre-cti charge injection images.
ci_pre_ctis = list(map(lambda ci_pattern :
                       ci_pattern.simulate_ci_pre_cti(shape=shape),
                       ci_patterns))

# Use every ci pattern to simulate a ci image.
datas = list(map(lambda ci_pre_cti, ci_pattern:
                    ci_data.simulate(ci_pre_cti=ci_pre_cti, frame_geometry=frame_geometry,
                                     ci_pattern=ci_pattern, cti_settings=cti_settings, cti_params=cti_params,
                                     read_noise=4.0, use_parallel_poisson_densities=True),
                 ci_pre_ctis, ci_patterns))

# Now, output every image to the data folder as the filename 'ci_data_normalization.fits'
list(map(lambda ci_data, normalization :
         util.numpy_array_2d_to_fits(array_2d=ci_data.image,
                                     file_path=ci_data_path + 'image_' + str(int(normalization)) + '.fits',
                                     overwrite=True),
         datas, normalizations))

# Output every pre-cti image to the data folder as the filename 'ci_pre_cti_normalization.fits'. This allows the
# calibration pipeline to load these images as the model pre-cti images, which is necessary for non-uniform ci patterns.
list(map(lambda ci_data, normalization :
         util.numpy_array_2d_to_fits(array_2d=ci_data.ci_pre_cti,
                                     file_path=ci_data_path + 'ci_pre_cti_' + str(int(normalization)) + '.fits',
                                     overwrite=True),
         datas, normalizations))


import matplotlib.pyplot as plt

from autocti import exc
from autocti.plotters import plotter_util


def line_from_line_region_and_arrays(line_region, array, mask, ci_frame):
    """

    Parameters
    -----------
    ci_frame : charge_injection.ci_frame.ChInj
    """
    if line_region is 'parallel_front_edge':
        return ci_frame.parallel_front_edge_line_binned_over_columns_from_frame(array=array, mask=mask)
    elif line_region is 'parallel_trails':
        return ci_frame.parallel_trails_line_binned_over_columns_from_frame(array=array, mask=mask)
    elif line_region is 'serial_front_edge':
        return ci_frame.serial_front_edge_line_binned_over_rows_from_frame(array=array, mask=mask)
    elif line_region is 'serial_trails':
        return ci_frame.serial_trails_line_binned_over_rows_from_frame(array=array, mask=mask)
    else:
        raise exc.PlottingException('The line region specified for the plotting of a line was invalid')

def plot_line_from_array_and_ci_frame(
        array, line_region, ci_frame, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='Stack', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='line'):

    line = line_from_line_region_and_arrays(line_region=line_region, array=array, mask=mask, ci_frame=ci_frame)

    plot_line(line=line, as_subplot=as_subplot,
               figsize=figsize,
               title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
               output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_line(line, as_subplot=False,
               figsize=(7, 7),
               title='Stack', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
               output_path=None, output_format='show', output_filename='line'):
    """Plot an array of hyper as a figure.

    Parameters
    -----------
    line : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    mask : ndarray of data.mask.Mask
        The masks applied to the hyper, the edge of which is plotted as a set of points over the plotted array.
    extract_line_from_mask : bool
        The plotter array is extracted using the mask, such that masked values are plotted as zeros. This ensures \
        bright features outside the mask do not impact the color map of the plot.
    as_subplot : bool
        Whether the array is plotted as part of a subplot, in which case the grid figure is not opened / closed.
    figsize : (int, int)
        The size of the figure in (rows, columns).
    aspect : str
        The aspect ratio of the hyper, specifically whether it is forced to be square ('equal') or adapts its size to \
        the figure size ('auto').
    cb_ticksize : int
        The size of the tick labels on the colorbar.
    cb_fraction : float
        The fraction of the figure that the colorbar takes up, which resizes the colorbar relative to the figure.
    cb_pad : float
        Pads the color bar in the figure, which resizes the colorbar relative to the figure.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    output_path : str
        The path on the hard-disk where the figure is output.
    output_filename : str
        The filename of the figure that is output.
    output_format : str
        The format the figue is output:
        'show' - display on computer screen.
        'png' - output to hard-disk as a png.
        'fits' - output to hard-disk as a fits file.'
    """

    if line is None:
        return

    plot_figure(line=line, as_subplot=as_subplot, figsize=figsize)

    plotter_util.set_title(title=title, titlesize=titlesize)
    set_xy_labels_and_ticksize(xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize)
    plotter_util.output_figure(line, as_subplot=as_subplot, output_path=output_path, output_filename=output_filename,
                               output_format=output_format)
    plotter_util.close_figure(as_subplot=as_subplot)


def plot_figure(line, as_subplot, figsize):
    """Open a matplotlib figure and plot the array of hyper on it.

    Parameters
    -----------
    line : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    as_subplot : bool
        Whether the array is plotted as part of a subplot, in which case the grid figure is not opened / closed.
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float or None
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    figsize : (int, int)
        The size of the figure in (rows, columns).
    aspect : str
        The aspect ratio of the hyper, specifically whether it is forced to be square ('equal') or adapts its size to \
        the figure size ('auto').
    cmap : str
        The colormap the array is plotted using, which may be chosen from the standard matplotlib colormaps.
    norm : str
        The normalization of the colormap used to plot the hyper, specifically whether it is linear ('linear'), log \
        ('log') or a symmetric log normalization ('symmetric_log').
    norm_min : float or None
        The minimum array value the colormap map spans (all values below this value are plotted the same color).
    norm_max : float or None
        The maximum array value the colormap map spans (all values above this value are plotted the same color).
    linthresh : float
        For the 'symmetric_log' colormap normalization ,this specifies the range of values within which the colormap \
        is linear.
    linscale : float
        For the 'symmetric_log' colormap normalization, this allowws the linear range set by linthresh to be stretched \
        relative to the logarithmic range.
    xticks_manual :  [] or None
        If input, the xticks do not use the array's default xticks but instead overwrite them as these values.
    yticks_manual :  [] or None
        If input, the yticks do not use the array's default yticks but instead overwrite them as these values.
    """
    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)
    plt.plot(line)

def set_xy_labels_and_ticksize(xlabelsize, ylabelsize, xyticksize):
    """Set the x and y labels of the figure, and set the fontsize of those labels.

    The x and y labels are always the distance scales, thus the labels are either arc-seconds or kpc and depend on the \
    units the figure is plotted in.

    Parameters
    -----------
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    """
    plt.xlabel('x (pixels)', fontsize=xlabelsize)
    plt.ylabel('y (pixels)', fontsize=ylabelsize)
    plt.tick_params(labelsize=xyticksize)


from autofit import conf
import os

from autofit.tools import path_util
from autocti.data import mask as msk
from autocti.charge_injection import ci_pattern
from autocti.charge_injection import ci_data

from autocti.model import arctic_settings

from workspace_jam.tools.data_makers import tools

from multiprocessing import Pool

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = workspace_path + '../../outputs/PyAutoCTI/'
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=output_path)

# Welcome to the pipeline runner. This tool allows you to CTI calibration data on strong lenses, and pass it to
# pipelines for a PyAutoCTI analysis. To show you around, we'll load up some example data and run it through some of
# the example pipelines that come distributed with PyAutoCTI.

# The runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out data-set names
# and pipelines to perform different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# The pipeline runner is fairly self explanatory. Make sure to checkout the pipelines in the
#  workspace/pipelines/examples/ folder - they come with detailed descriptions of what they do. I hope that you'll
# expand on them for your own personal scientific needs

# It is convenient to specify the image type and model used to simulate that image as strings, so that if the
# pipeline is applied to multiple images we don't have to change all of the path entries in the
# load_ci_data_from_fits function below.

ci_data_type = 'ci_lines_uniform' # Charge injection data consisting of 2 images with uniform injections.
ci_data_model = 'parallel_x3' # Shows the data was creating using a parallel CTI model with one species.

# Create the path where the data will be output, which in this case is
# '/workspace/data/ci_lines_uniform/parallel_x2_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=['data', ci_data_type, ci_data_model])

# This tools loads the correct shape of the line and frame geometry, for the ci_data_resoultion string.
shape, frame_geometry = tools.charge_line_shape_and_frame_geometry_from_direction(direction='parallel')

# Specify the charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, 30, 0, 1), (330, 360, 0, 1), (660, 690, 0, 1), (990, 1020, 0, 1),
              (1320, 1350, 0, 1), (1650, 1680, 0, 1), (1980, 2010, 0, 1)]

# The normalization of the charge injection for each image.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# Create the charge injection pattern objects used for this pipeline.
patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

# There are 8 images to load, which is what the normalizations list tells us. To load them, its easiest to create a
# 'data' list and then iterate over a for loop, to append each set of data to the list of data we pass to the pipeline
# when we run it.

datas = []

for index, normalization in enumerate(normalizations):

    datas.append(ci_data.ci_data_from_fits(
                 frame_geometry=frame_geometry, ci_pattern=patterns[index],
                 image_path=ci_data_path + 'line_' + str(int(normalization)) + '.fits',
                 ci_pre_cti_path=ci_data_path +'/ci_pre_cti_' + str(int(normalization)) + '.fits',
                 noise_map_from_single_value=0.1))

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For parallel ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
parallel_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=True, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(parallel=parallel_cti_settings)

def mask_function_for_bug(shape, ci_frame):
    return msk.Mask.from_masked_regions(shape=shape, masked_regions=[(0, 330, 0, 1)])

# Running a pipeline is easy, we simply import it from the pipelines folder and pass the ci data to its run function.
# Below, we'll' use a 2 phase example pipeline to fit the data with a one species parallel CTI model.
# Checkout workspace/pipelines/examples/parallel_x1_species.py' for a full description of the pipeline.

# The pool command tells our code to parallelize the analysis over 2 CPU's, where each CPU fits a different charge
# injection image

from workspace_jam.pipelines.initialize import parallel_x3

pipeline_init = parallel_x3.make_pipeline(phase_folders=[ci_data_type, ci_data_model],
                                          mask_function=mask_function_for_bug)

from workspace_jam.pipelines.normal.from_initialize import parallel_x3

pipeline_normal = parallel_x3.make_pipeline(phase_folders=[ci_data_type, ci_data_model],
                                            mask_function=mask_function_for_bug)

pipeline = pipeline_init + pipeline_normal

pipeline.run(ci_datas=datas, cti_settings=cti_settings, pool=Pool(processes=2),
             assert_optimizer_pickle_matches=False)


from matplotlib import pyplot as plt
from autocti.plotters import line_plotters, array_plotters
from autocti.plotters import plotter_util

def plot_image(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_image'):
    """Plot the observed image of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : CIFrame
        The image of the data.
    """

    array_plotters.plot_array(
        array=fit.image, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_noise_map(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Noise-Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_noise_map'):
    """Plot the observed noise_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    noise_map : CIFrame
        The noise_map of the data.
    """

    array_plotters.plot_array(
        array=fit.noise_map, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_signal_to_noise_map(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Signal-to-Noise Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_signal_to_noise_map'):
    """Plot the observed signal_to_noise_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : CIFrame
        The signal_to_noise_map of the data.
    """

    array_plotters.plot_array(
        array=fit.signal_to_noise_map, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_ci_pre_cti(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Pre-CTI CI Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_ci_pre_cti'):
    """Plot the observed ci_pre_cti of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the data.
    """

    array_plotters.plot_array(
        array=fit.ci_pre_cti, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_ci_post_cti(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Post-CTI CI Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_ci_post_cti'):
    """Plot the observed ci_post_cti of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_post_cti : CIFrame
        The ci_post_cti of the data.
    """

    array_plotters.plot_array(
        array=fit.ci_post_cti, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_residual_map(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Residual Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_residual_map'):
    """Plot the observed residual_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    residual_map : CIFrame
        The residual_map of the data.
    """

    array_plotters.plot_array(
        array=fit.residual_map, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_chi_squared_map(
        fit, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Chi-Squared Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_chi_squared_map'):
    """Plot the observed chi_squared_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : CIFrame
        The chi_squared_map of the data.
    """

    array_plotters.plot_array(
        array=fit.chi_squared_map, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_noise_scaling_maps(
        fit_hyper, mask=None, extract_array_from_mask=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Fit Chi-Squared Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_noise_scaling_maps'):
    """Plot the observed chi_squared_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : CIFrame
        The chi_squared_map of the data.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=len(fit_hyper.noise_scaling_maps))

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for index in range(len(fit_hyper.noise_scaling_maps)):

        plt.subplot(rows, columns, index+1)

        array_plotters.plot_array(
            array=fit_hyper.noise_scaling_maps[index], mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
            figsize=figsize, aspect=aspect,
            cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
            cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
            cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
            title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()

def plot_image_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_image_line'):
    """Plot the observed image of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : CIFrame
        The image of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.image, line_region=line_region, ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)
    

def plot_noise_map_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='Noise Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_noise_map_line'):
    """Plot the observed noise_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    noise_map : CIFrame
        The noise_map of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.noise_map, line_region=line_region, ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)
    
    
def plot_signal_to_noise_map_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='Signal-To-Noise Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_signal_to_noise_map_line'):
    """Plot the observed signal_to_noise_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : CIFrame
        The signal_to_noise_map of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.signal_to_noise_map, line_region=line_region,  ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)
    
    
def plot_ci_pre_cti_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='CI Pre-CTI Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_ci_pre_cti_line'):
    """Plot the observed ci_pre_cti of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.ci_pre_cti, line_region=line_region,  ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)
    
    
def plot_ci_post_cti_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='CI Post-CTI Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_ci_post_cti_line'):
    """Plot the observed ci_post_cti of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_post_cti : CIFrame
        The ci_post_cti of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.ci_post_cti, line_region=line_region, ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)
    
    
def plot_residual_map_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='Residual Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_residual_map_line'):
    """Plot the observed residual_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    residual_map : CIFrame
        The residual_map of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.residual_map, line_region=line_region, ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)
    
    
def plot_chi_squared_map_line(
        fit, line_region, mask=None, as_subplot=False,
        figsize=(7, 7),
        title='Chi-Squared Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='fit_chi_squared_map_line'):
    """Plot the observed chi_squared_map of the ccd data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : CIFrame
        The chi_squared_map of the data.
    """
    line_plotters.plot_line_from_array_and_ci_frame(
        array=fit.chi_squared_map, line_region=line_region, ci_frame=fit.ci_data_fit.ci_frame, mask=mask, as_subplot=as_subplot,
       figsize=figsize,
       title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
       output_path=output_path, output_format=output_format, output_filename=output_filename)